=begin

Copyright 2012, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         ae_Snapshot/Snapshot.rb
Author:       Andreas Eisenbarth
Description:  This plugin allows to make customizable buttons for creating snapshots of SketchUp files.
Usage:        menu View → Toolbars → Snapshot
Version:      0.3.1
Date:         07.10.2012

=end


require 'sketchup.rb'


module AE


module Snapshot

  @default_options = {
    "0" => {
      "name" => "+",
      "description" => "Edit to customize your snapshot tool...",
      "icon" => "./snapshot_shoot_24.png",
      "autosave" => false,        # whether to save automatically when the skp is saved
      "location-same" => true,
      "location-custom" => false,
      "location" => ENV["HOME"]||ENV["HOMEPATH"],  # custom location
      "location-shortcut" => false,
      "filename-format" => "{model}_{nr}",
      "version-format" => "%d-%d-%d",        # printf format: "%3d" gives minimum width of 3 digits with trailing spaces if necessary, %03d" gives minimum width of 3 digits with trailing 0
      "version-part" => 1,       # the part of the version number to increase (major, minor, revision number)
      "version-increment" => 1,  # the increment by which the version number is changed (-1, +1, +2...)
      "version-change" => true,
      "date-format" => "YYYY-MM-DD_hh-mm",
      "filename-ask" => false,
      "fileformat" => "png",
      "alpha" => true,
      "width" => 512,
      "height" => 384,
    }
  }


  @options = (eval(Sketchup.read_default("Plugins_ae", "Snapshot", "{}")) || {}).merge(@default_options) rescue @default_options
  @options = @default_options if @options["toolbar-plus"] # TEMPORARY: reset options from Snapshot.rb version 0.1
  def self.options=(opt)
    @options = (opt.is_a?(Hash) ? opt : {}).merge(@default_options)
    Sketchup.write_default("Plugins_ae", "Snapshot", @options.inspect.inspect[1..-2])
  end
  def self.options
    return @options
  end

  @formats = ["png", "jpg", "jpeg", "bmp", "tiff", "tif"]
  @tmp_dir = File.expand_path(
    [ENV['TMPDIR'], ENV['TMP'], ENV['TEMP'], ENV['USERPROFILE'], '/tmp', '.'].inject(nil){|t,dir|
      (!t && dir && File.directory?(dir) && File.writable?(dir))? dir : t
    }
  )
  @dir = File.dirname(__FILE__)


  class OpenObserver < Sketchup::AppObserver
    def initialize(options)
      @options = options
    end
    # Observe when model is saved.
    def onOpenModel(model)
      model.add_observer(SaveObserver.new(@options))
    end
    alias_method(:onNewModel, :onOpenModel)
  end


  class SaveObserver < Sketchup::ModelObserver
    def initialize(options)
      @options = options
    end
    # Create a snapshot after saving the model.
    def onPostSaveModel(model)
      t = UI.start_timer(1, false){ UI.stop_timer(t)
        Snapshot.save(@options)
      }
    end
  end


  # Save a snapshot of the current view.
  #
  # @param [Hash] options a hash of options
  #
  def self.save(options)
    filepath = self.get_filepath(options) || return
    if options["filename-ask"]
      filepath = UI.savepanel("Save a snapshot", filepath) || return
      options["location"] = File.dirname(filepath)
    end
    m = Sketchup.active_model
    r = m.rendering_options
    alpha = options["alpha"] && options["fileformat"][/png|tif/]
    # Disable ground and horizon for transparent export.
    if alpha
      drawGround, r["DrawGround"] = r["DrawGround"], false
      drawHorizon, r["DrawHorizon"] = r["DrawHorizon"], false
    end
    keys = {
      :filename => filepath,
      :width => options["width"], # TODO: preserve aspect ratio (?)
      :height => options["height"],
      :antialias => true,
      :compression => 0.9, # Actually only for jpeg, not used here.
      :transparent => alpha
    }
    m.active_view.write_image(keys)
    Sketchup.status_text = "A snapshot has been saved as #{filepath}"
    # Reset ground and horizon.
    (r["DrawGround"] = drawGround; r["DrawHorizon"] = drawHorizon) if alpha
    # Optionally create shortcut from image location (*.lnk file) to model file.
    self.create_shortcut(filepath, m.path) if options["location-shortcut"]
  end


  # Get the filepath according to the settings.
  #
  # @private
  def self.get_filepath(options)
    path = Sketchup.active_model.path
    return (UI.messagebox("You need to save the model first."); false) if path.empty?
    location = (options["location-same"])? File.dirname(path) : options["location"]
    location = File.expand_path(location)
    # model name
    model = File.basename(path, ".skp")
    # date
    t = Time.now.to_a
    date = options["date-format"].
      sub(/(Y+)/, sprintf("%d",t[5])).
      sub(/M+/i, sprintf("%02d",t[4])).
      sub(/D+/, sprintf("%02d",t[3])).
      sub(/h+/, sprintf("%02d",t[2])).
      sub(/m+/i, sprintf("%02d",t[1])).
      sub(/s+/, sprintf("%02d",t[0]))
    # version number
    path = File.join(location, options["filename-format"] + options["fileformat"])
    nr = (options["filename-format"][/\{nr\}/])? self.get_increment(path, options) : ""
    # full file name
    filename = options["filename-format"].sub(/\{model\}/, model).sub(/\{date\}/, date).sub(/\{nr\}/, nr) + "." + options["fileformat"]
    return File.join(location, filename)
  end


  # Get the version number for the new snapshot.
  #
  # @param [options] hash of the options
  # @param [String] path the path where to look for previous versions
  #
  def self.get_increment(path, options)
    # check for previous snapshots
    previous = Dir.glob( path.sub(/\{nr\}/, "*").sub(/\{date\}/, "*").sub(/\..{3,4}$/, ".{#{@formats.join(',')}}") )
    digits = options["version-format"].scan(/\%.*?d/).length
    nr = [0]*digits
    if previous.empty?
      nr[0] = 1
    else
      # There exist already snapshots.
      # Find the one with the highest number.
      nr_regexp = options["version-format"].gsub(/\%.*?d/, "([0-9]+)?").gsub(/9\]\+\)\?(.*?)\(\[0/, "9]+)?(#{Regexp.quote("\1")}[0")
      #format_regexp = "\\.(?:#{@formats.join('|')})"
      regexp = Regexp.new( Regexp.quote(File.basename(path)).sub("\\{nr\\}", nr_regexp).sub(/\\\..{3,4}$/, "") )
      previous.collect{|p|
        p.scan(regexp).first.compact.collect{|c| c.to_i}
      }.max.each_with_index{|c,i| nr[i] = c}
      # add the increment(s)
      if options["version-change"]
        nr[options["version-part"].to_i-1] = (nr[options["version-part"].to_i-1]||0) + options["version-increment"].to_i
      end
    end
    # return the number
    return sprintf(options["version-format"], *nr)
  end


  # Display the settings dialog.
  #
  def self.settings_dialog
    @dlg = UI::WebDialog.new("Snapshot settings", false, false, 300, 300, 500, 400, true)
    @dlg.set_size(500, 400) # default size, will be adjusted later
    @dlg.set_position(300, 300)
    html_path = File.join(@dir, "settings_dialog.htm")
    @dlg.set_file(html_path)
    # callbacks
    # init: pass the default options to the form
    @dlg.add_action_callback("init") {|dlg, param|
      dlg.execute_script("document.getElementsByTagName('body')[0].style.background='"+dlg.get_default_dialog_color+"'")
      dlg.execute_script("start(" + @options.inspect.gsub(/=>/,":") + ")")
    }
    # receives the document's inner dimensions in order to obtain the window border size
    @dlg.add_action_callback("windowInner") {|dlg, params|
      params = eval(params)
      @windowBorderWidth = (500-params[0].to_f)/2 if !@windowBorderWidth
      @windowTitlebarHeight = 400-params[1].to_f if !@windowTitlebarHeight
      dlg.execute_script("resizeHeight()")
    }
    # receives the html's dimensions in order set a specific webdialog size
    @dlg.add_action_callback("resizeHeight") {|dlg, params|
      params = eval(params)
      width = 2*@windowBorderWidth+params[0].to_f
      height = @windowTitlebarHeight+params[1].to_f
      dlg.set_size(width, height)
    }
    # Select file # native file input gives security violation when trying to set its value. Also this way we are sure not to get a URI with file:///.
    @dlg.add_action_callback("selectFile") { |dlg, param|
      location, receiver = *param.split("|")
      location.sub!(/^\.(?=\\|\/)/, @dir)
      file = UI.openpanel("Select a file", location) || return
      file = File.expand_path(file).sub(@dir, ".")
      dlg.execute_script("try{"+receiver+"='#{file.inspect[1..-2]}'}catch(e){}")
    }
    # Select folder, this is not possible in the WebDialog.
    @dlg.add_action_callback("selectFolder") { |dlg, param|
      location, receiver = *param.split("|")
      location.sub!(/^\.(?=\\|\/)/, @dir)
      file = UI.openpanel("Select a file in the folder", location) || return
      folder = File.dirname(File.expand_path(file)).sub(@dir, ".")
      dlg.execute_script("try{"+receiver+"='#{folder.inspect[1..-2]}'}catch(e){}")
    }
    # Apply button
    @dlg.add_action_callback("receiveOptions") { |dlg, param|
      @options = eval(dlg.get_element_value("results")) rescue @options
      Sketchup.write_default("Plugins_ae", "Snapshot", @options.inspect.inspect[1..-2])
      # Update observers if autosave setting has changed
      @options.each{|id, options|
        @toolbarbuttons << self.toolbar_add(@toolbar, options) unless @toolbarbuttons.include?(options["name"])
        if options.is_a?(Hash) && options["autosave"]
          Sketchup.add_observer(@openobservers[id] = OpenObserver.new(options))
          Sketchup.active_model.add_observer(@saveobservers[id] = SaveObserver.new(options))
        else
          Sketchup.active_model.remove_observer(@openobservers[id])
          Sketchup.active_model.remove_observer(@saveobservers[id])
          @openobservers[id], @saveobservers[id] = []
        end
      }
      dlg.close
    }
    # Cancel button
    @dlg.add_action_callback("cancel") { |dlg, param| dlg.close}
    # for debugging
    @dlg.add_action_callback("puts") { |dlg, param|
      puts(param)
    }
    # show the webdialog
    @dlg.show_modal
  end



# Creates a Windows shortcut.
# @param [String] source the filepath of the shortcut
# @param [String] target the filepath to link to
#
def self.create_shortcut(source, target)
    # set source's extension to lnk
    source.sub!(/(\..{0,4})?$/, ".lnk")
    # create a Visual Basic file to launch the batch script without black console window
    vbs = File.join(@tmp_dir, "create_shortcut.vbs")
    File.open(vbs, "w"){|f|
      f.puts %[set WshShell = CreateObject("WScript.Shell")]
      f.puts %[set oShellLink = WshShell.CreateShortcut(#{source.inspect})]
      f.puts %[oShellLink.TargetPath = #{target.inspect}]
      f.puts %[oShellLink.WindowStyle = 1]
      f.puts %[oShellLink.Save]
      f.puts %[WScript.Quit]
    }
    system(%[wscript #{vbs}])
end # def run_shell_command



# Adds a button to a toolbar.
# @param [UI::Toolbar] toolbar
# @param [Hash] options
#
def self.toolbar_add(toolbar, options)
  cmd = UI::Command.new(options["name"]){ self.save(options) }
  large_icon = options["icon"].to_s.sub(/^\.(?=\\|\/)/, @dir)
  small_icon = large_icon.sub(/24(?=.*?$)/,"16"); small_icon = large_icon unless File.exists?(small_icon)
  cmd.small_icon = small_icon
  cmd.large_icon = large_icon
  cmd.tooltip = options["description"]
  toolbar.add_item(cmd)
  return options["name"]
end # def run_shell_command



  # Initialize the UI.
  unless file_loaded?(__FILE__)

    # Toolbar
    @toolbar = UI::Toolbar.new("Snapshot")
    @openobservers ||= {}
    @saveobservers ||= {}
    @toolbarbuttons ||= []

    # Custom commands
    @options.each{|id, options|
      next if !options.is_a?(Hash) || options["name"] == "+"
      @toolbarbuttons << self.toolbar_add(@toolbar, options)
      # Observers
      if options["autosave"]
        Sketchup.add_observer(@openobservers[id] = OpenObserver.new(options))
        Sketchup.active_model.add_observer(@saveobservers[id] = SaveObserver.new(options))
      end
    }

    # Settings command
    cmd_settings = UI::Command.new("Show the settings dialog."){ self.settings_dialog }
    cmd_settings.small_icon = File.join(@dir,"snapshot_settings_16.png")
    cmd_settings.large_icon = File.join(@dir,"snapshot_settings_24.png")
    cmd_settings.tooltip = "Show a dialog to set settings of the Snapshot function."
    @toolbar.add_item(cmd_settings)

    @toolbar.show if @toolbar.get_last_state == 1
    # TODO: UI::Toolbar.restore ?
  end

end # module Snapshot



end # module AE



file_loaded(__FILE__)
